# References
#
# The first reference is for the DA algorithm for one-to-one matching problems.
# The second reference is for making file dialogs.
#
# Author: Cristi Burca
# Date(accessed): 2019/11/30
# Code version: <code version>
#Availability: https://gist.github.com/scribu/104ec4ba54207db8c6e8
#
# Author: chanmaru
# Date(accessed): 2020/3/21
# Availability: https://qiita.com/chanmaru/items/1b64aa91dcd45ad91540


from collections import defaultdict
import pandas as pd
import numpy as np
import math, codecs, os, tkinter, tkinter.filedialog, tkinter.messagebox

# return a student who has not been matched yet.
def student_without_match(matches, students):
    for s in students:
        if s not in matches:
            return s


#
# Deferred Acceptance (Gale-Shapley) Algorithm
#
def deferred_acceptance(student_prefs, college_prefs, quotas={}):
    college_queue = defaultdict(int)

    students = list(student_prefs.keys())

    matches = {}

    while True:
        stu = student_without_match(matches, students)
        if not stu:
            break

        college_index = college_queue[stu]
        college_queue[stu] += 1

        try:
            col = student_prefs[stu][college_index]
        except IndexError:
            matches[stu] = stu
            continue

#        print('Trying %s with %s... ' % (stu, col), end='')

        # get the least preferred student for col.
        # worst_stu is None if a seat is available.
        prev_stu = matches.get(col, None)
        q = quotas.get(col, 1)  # quota is 1 if not specified.
        if q==0:
#            print('rejected')
            continue
        elif not prev_stu or q > len(prev_stu):
            worst_stu = None
        else:
            worst_stu = prev_stu[0]
            for s in prev_stu:
                if college_prefs[col].index(s) > \
                 college_prefs[col].index(worst_stu):
                    worst_stu = s

        # reject if the student is not acceptable.
        print(stu, col)
        if not stu in college_prefs[col]:
            pass
#            print('rejected')

        # match if there is an available seat.
        elif not worst_stu:
            matches[stu] = col
            if col in matches:
                matches[col].append(stu)
            else:
                matches[col] = [stu]
#            print('matched')

        # match if the student is preferred to an existing one.
        # reject otherwise.
        elif college_prefs[col].index(stu) < \
             college_prefs[col].index(worst_stu):
            matches[stu] = col
            matches[col].remove(worst_stu)
            matches[col].append(stu)

            del matches[worst_stu]

#            print('matched')
        else:
            pass
#            print('rejected')

    return {stu: matches[stu] for stu in student_prefs.keys()}


#
# Controlled choice by Abudulkadiroglu and Sonmez (2003)
#
def DA_type_quota(student_prefs, college_prefs, types={}, quota={}):

    students = list(student_prefs.keys())
    colleges = list(college_prefs.keys())

    college_prefs_ext = {}
    quota_ext = {}

    student_prefs_ext = student_prefs.copy()

    for stu in students:
        types[stu] = types.get(stu,1)

    stu_type1 = [k for k, v in types.items() if v == 1]
    stu_type2 = [k for k, v in types.items() if v == 2]

    for col in colleges:
        # college names for the extended preferences
        col_ext = [col+'_0', col+'_1', col+'_2']

        q = quota.get(col,[1,1,1])
        quota_ext[col_ext[0]] = q[1]+q[2]-q[0]
        quota_ext[col_ext[1]] = q[0]-q[2]
        quota_ext[col_ext[2]] = q[0]-q[1]

        col_pref_type1 = college_prefs[col].copy()
        col_pref_type2 = college_prefs[col].copy()

        for s_type1 in stu_type1:
            if s_type1 in col_pref_type2:
                col_pref_type2.remove(s_type1)
        for s_type2 in stu_type2:
            if s_type2 in col_pref_type1:
                col_pref_type1.remove(s_type2)
        college_prefs_ext[col_ext[0]] = college_prefs[col]
        college_prefs_ext[col_ext[1]] = col_pref_type1
        college_prefs_ext[col_ext[2]] = col_pref_type2

        for stu in students:
            stu_pref = student_prefs_ext[stu].copy()
            if col in stu_pref:
                i = stu_pref.index(col)
                stu_pref.remove(col)
                for j in range(3):
                    stu_pref.insert(i,col_ext[(3-j)%3])
            student_prefs_ext[stu] = stu_pref

    matches_ext = \
        deferred_acceptance(student_prefs_ext, college_prefs_ext,quota_ext)
    return {stu: matches_ext[stu] if stu == matches_ext[stu] \
            else matches_ext[stu][:-2] \
            for stu in student_prefs.keys()}


#### Main Program ####

## setting for reading the csv files ##
file_faculty_def = 'test_faculty.csv'
file_students_def = 'test_students.csv'
file_matching = 'DA_output.csv'

# ファイル選択ダイアログの表示
root = tkinter.Tk()
root.withdraw()
fTyp = [("","*.csv")]
iDir = os.path.abspath(os.path.dirname(__file__))
tkinter.messagebox.showinfo('DAマッチングプログラム',\
    '最初に教員の選好情報ファイル\n続いて学生の選好情報ファイルを選択してください！')
f1 = tkinter.filedialog.askopenfilename(filetypes = fTyp,initialdir = iDir, \
        title="教員の選好情報ファイルを選択")
file_faculty = os.path.join(iDir,file_faculty_def) if f1 == '' else f1

f2 = tkinter.filedialog.askopenfilename(filetypes = fTyp,initialdir = iDir, \
        title="学生の選好情報ファイルを選択")
file_students = os.path.join(iDir,file_students_def) if f2 == '' else f2

f_id_index = 0      # index for id in faculty csv.
f_q0_index = 2      # index for Zentai quota
f_q1_index = 3      # index for City and Planning (Toshi) quota
f_prefs_index = 5   # index of the beginning of preferences in faculty csv.
s_id_index = 0      # index for id in student csv.
s_type_index = 1    # index for type in student csv.
s_prefs_index = 2   # index of the beginning of preferences in student csv.
city_type_csv = '都市'   # specify the value for 'Toshi' type in student csv.
## end of setting ##

faculty_prefs = {}
student_prefs = {}
types = {}
quotas = {}
city_type_for_da = 2
other_type_for_da = 1

# read csv files
with codecs.open(file_faculty, "r", "Shift-JIS", "ignore") as file:
    ff = pd.read_csv(file)
    print(ff)
with codecs.open(file_students, "r", "Shift-JIS", "ignore") as file:
    fs = pd.read_csv(file)
    print(fs)
#ff = pd.read_csv(file_faculty)
#fs = pd.read_csv(file_students)

# convert to list objects
listf = ff.values.tolist()
lists = fs.values.tolist()

# make faculty preferences and quotas
for item in listf:
    faculty_prefs[item[f_id_index]] = \
        [i for i in item[f_prefs_index:] if type(i)==str and i.strip() !='']
    quotas[item[f_id_index]] = \
        [item[f_q0_index],item[f_q0_index],item[f_q0_index]]
#       If you use the type-specific quota for "city_type_csv" type, use
#        [item[f_q0_index],item[f_q0_index],item[f_q1_index]]

print(faculty_prefs)
print(quotas)

# make student preferences and types
for item in lists:
    student_prefs[item[s_id_index]] = \
        [i for i in item[s_prefs_index:] if type(i)==str and i.strip() !='']
    if item[s_type_index] == city_type_csv:
        types[item[s_id_index]] = city_type_for_da
    else:
        types[item[s_id_index]] = other_type_for_da

#print(student_prefs)

# compute DA with controlled choice
# matches: dictionary object (key: student id, value: matched faculty id)
matches = DA_type_quota(student_prefs, faculty_prefs, types, quotas)


# matches_f: dictionary object (key: faculty id, value: list of students)
matches_f = {}
for s in matches.keys():
    if matches[s] == s:
        continue
    elif not matches[s] in matches_f:
        matches_f[matches[s]] = [s]
    else:
        matches_f[matches[s]].append(s)

print(os.path.dirname(file_faculty))
file_out = os.path.join(os.path.dirname(file_faculty), file_matching)
# Print output
print(sorted(matches_f.items(), key=lambda x: x[0]))
with open(file_out, 'wt') as f:
    for k, v in matches_f.items():
        liststu = ','.join(v)
        f.write(k+','+liststu+'\n')
